from whylogs.datasets.ecommerce import Ecommerce
from whylogs.datasets.employee import Employee
from whylogs.datasets.weather import Weather

__ALL__ = [
    # column
    Weather,
    Ecommerce,
    Employee,
]
